structure KFelhok :> KFelhok =
    (* dp04s nhf v1.2 -- 2004. mrcius 26. Benk Tams, Szeredi Pter *)
struct

(* A bemenetet, ill. a kimenetet a terminlra irnythatjuk Unix alatt
   a /dev/stdin, ill. a /dev/stdout, MS-DOS alatt mindkt esetben a con
   llomnynv megadsval.
*)

exception KFelhok of string

fun lexikogr_hasonlit ((x1,y1), (x2,y2)) =
    if      x1 < x2 orelse x1 = x2 andalso y1 < y2 then LESS
    else if x1 > x2 orelse x1 = x2 andalso y1 > y2 then GREATER
	 else EQUAL


(* helyes_feladvany f = az f feladvny helyessge
   helyes_feladvany : TFelhok.radarkep -> bool
*)
fun helyes_feladvany ((SorOsszegek, OszlopOsszegek, RefPL) : TFelhok.radarkep) =
    let
        val n = length SorOsszegek
        val m = length OszlopOsszegek

	fun ref_pont_koord (TFelhok.f(Sor, Oszlop)) = (Sor, Oszlop)
          | ref_pont_koord (TFelhok.e(Sor, Oszlop)) = (Sor, Oszlop)

	fun helyes_koord (Sor, Oszlop) =
	    Sor > 0 andalso Sor <= n andalso
	    Oszlop > 0 andalso Oszlop <= m 

	val helyes_ref_pont = helyes_koord o ref_pont_koord

	fun helyes_osszegek ((osszeg::osszegek), max) =
	    osszeg > ~2 andalso osszeg <= max andalso
	    helyes_osszegek (osszegek, max)
          | helyes_osszegek _ = true

        fun lexikografikusan_hasonlit (rp1, rp2) =
	    lexikogr_hasonlit (ref_pont_koord rp1, ref_pont_koord rp2)

	fun szomszedok_kulonbozok (rp1::(rps as rp2::_)) =
	    lexikografikusan_hasonlit (rp1, rp2) <> EQUAL
	    andalso szomszedok_kulonbozok rps
	  | szomszedok_kulonbozok _ = true
    in
	(* az sszegek a megadott intervallumon bell vannak: *)
	helyes_osszegek (SorOsszegek, m) andalso
	helyes_osszegek (OszlopOsszegek, n) andalso
	(* a RefPL lista minden eleme helyes: *)
	(List.all (fn x => x) (map helyes_ref_pont RefPL)) andalso
	(* a RefPL lista lexikografikusan rendezett: *)
	Listsort.sorted lexikografikusan_hasonlit RefPL andalso
	(* a RefPL listban szerepl (Sor,Oszlop) koordintk
	 pronknt klnbznek (a rendezettsg miatt elegend
	 ezt a szomszdos elemekre kiktni): *)
	szomszedok_kulonbozok RefPL
    end

(* felhok_be f = az f szveges llomnybl beolvasott feladvny.
   felhok_be : string -> TFelhok.radarkep
 *)
fun felhok_be f =
    let
	fun str2int s =
	    case Int.fromString s of
		NONE => raise KFelhok ("Nem vrt karaktersorozat: '" ^ s ^ "'.")
	      | SOME v => v
	val str2ints = map str2int o String.tokens Char.isSpace
	fun str2ref s =
	    case String.tokens Char.isSpace s of
		["f",y,x] => TFelhok.f (str2int y, str2int x)
	      | ["e",y,x] => TFelhok.e (str2int y, str2int x)
	      | _ => raise KFelhok ("Nem vrt karaktersorozat: '" ^ s ^ "'.")
	fun list2screen (rs::cs::refs) = (str2ints rs, str2ints cs, map str2ref refs)
	  | list2screen _ = raise KFelhok "A bemenetben nincs 2 nemres sor."
	val istr = TextIO.openIn f
	val all  = TextIO.inputAll istr before TextIO.closeIn istr
	val rkep = list2screen (String.tokens (fn c => c = #"\n") all)
    in
	if helyes_feladvany rkep then rkep
	else raise KFelhok ("Hibs feladvny a(z) " ^ f ^ " llomnyban.")
    end

(* pretty_print radarkep felhok = a kitlttt radarkep "csinos" formban
   pretty_print : TFelhok.radarkep * TFelhok.felhok -> string
 *)
fun pretty_print ((rss,css,_),cs) =
    let
	fun cloudy (y,x) (y1,x1,y2,x2) = y >= y1 andalso y <= y2 andalso x >= x1 andalso x <= x2
	fun mark k = if List.exists (cloudy k) cs then "#" else "."
	fun cols (_::xs,r,c,rs) = "| " ^ mark (r,c) ^ " " ^ cols (xs,r,c+1,rs)
	  | cols (_,_,_,rs) = "| " ^ Int.toString rs ^ "\n"
	fun separator (_::xs) = "+---" ^ separator xs
	  | separator _ = "+\n"
	fun rows (rs::rss,r) = cols (css,r,1,rs) ^ separator css ^ rows (rss,r+1)
	  | rows _ = "";
	fun colsum (cs::css) = (if cs > 99 then " " else "  ") ^ Int.toString cs
	    ^ (if cs < 0 orelse cs > 9 then "" else " ") ^ colsum css
	  | colsum _ = "\n"
    in
	separator css ^ rows (rss,1) ^ colsum css
    end

(* felhok_ki (m,t,fs) = a t feladvny fs megoldslistjt olvashat
   formban kirja az m szveges llomnyba.
   felhok_ki : string * TFelhok.radarkep * TFelhok.felhok list -> unit
*)
fun felhok_ki (m,t,fs) =
    let
	fun clouds2coords ((y1,x1,_,_),(y3,x3,_,_)) = ((y1,x1),(y3,x3))
	val cmpClouds = lexikogr_hasonlit o clouds2coords
	val sortClouds = Listsort.sort cmpClouds
	fun sortCloudsLists fs = 
	    let fun cmpLists cmpFun (c1::cs1,c2::cs2) = 
		    (case cmpFun (c1,c2) of
		        EQUAL => cmpLists cmpFun (cs1,cs2)
		      | res => res
		    )
		  | cmpLists _ (_::_,[]) = GREATER
		  | cmpLists _ ([],_::_) = LESS
		  | cmpLists _ ([],[])   = EQUAL
	    in Listsort.sort (cmpLists cmpClouds) (map sortClouds fs)
	    end
	fun pp (f,s) = s ^ (pretty_print (t,f) handle KFelhok s => s) ^ "\n"
	val ostr = TextIO.openOut m
	val _ = TextIO.output (ostr,foldl pp "" (sortCloudsLists fs))
    in
	TextIO.closeOut ostr
    end

(* megold (f,m) = beolvas egy feladvnyt az f szveges llomnybl s kirja az
                  sszes megoldst az m szveges llomnyba. Ehhez termszetesen
                  szksge van a Felhok.felhok fggvnyre. Eredmnye az f
                  nevet, a megoldsok szmt s a futsi idt msodpercben
                  tartalmaz fzr.
   megold : string * string -> string
*)
fun megold (f,m) =
    let val t = felhok_be f
	val start = Timer.startCPUTimer ()
	val fs = Felhok.felhok t
	val tim = #usr (Timer.checkCPUTimer start)
	val _ = felhok_ki (m,t,fs)
    in
	"Ler: " ^ f ^ ", " ^
	Int.toString (length fs) ^ " megolds, " ^
	"id: " ^ Time.fmt 2 tim ^ " mp\n"
    end
end
